@@ -6,7 +6,6 @@
 #include "alloc/backpointers.h"
 
 #include "btree/bbpos.h"
-#include "btree/bkey_buf.h"
 #include "btree/cache.h"
 #include "btree/update.h"
 #include "btree/interior.h"
@@ -187,7 +186,7 @@ static int bch2_backpointer_del(struct btree_trans *trans, struct bpos pos)
 
 static inline int bch2_backpointers_maybe_flush(struct btree_trans *trans,
 					 struct bkey_s_c visiting_k,
-					 struct bkey_buf *last_flushed)
+					 struct wb_maybe_flush *last_flushed)
 {
 	return !static_branch_unlikely(&bch2_backpointers_no_use_write_buffer)
 		? bch2_btree_write_buffer_maybe_flush(trans, visiting_k, last_flushed)
@@ -197,7 +196,7 @@ static inline int bch2_backpointers_maybe_flush(struct btree_trans *trans,
 static int backpointer_target_not_found(struct btree_trans *trans,
 				  struct bkey_s_c_backpointer bp,
 				  struct bkey_s_c target_k,
-				  struct bkey_buf *last_flushed,
+				  struct wb_maybe_flush *last_flushed,
 				  bool commit)
 {
 	struct bch_fs *c = trans->c;
@@ -260,7 +259,7 @@ fsck_err:
 static struct btree *__bch2_backpointer_get_node(struct btree_trans *trans,
 						 struct bkey_s_c_backpointer bp,
 						 struct btree_iter *iter,
-						 struct bkey_buf *last_flushed,
+						 struct wb_maybe_flush *last_flushed,
 						 bool commit)
 {
 	struct bch_fs *c = trans->c;
@@ -275,7 +274,7 @@ static struct btree *__bch2_backpointer_get_node(struct btree_trans *trans,
 				  0);
 	struct btree *b = bch2_btree_iter_peek_node(iter);
 	if (IS_ERR(b))
-		goto err;
+		return b;
 
 	if (!b) {
 		/* Backpointer for nonexistent tree depth: */
@@ -284,8 +283,7 @@ static struct btree *__bch2_backpointer_get_node(struct btree_trans *trans,
 		struct bkey_s_c k = { &iter->k };
 
 		int ret = backpointer_target_not_found(trans, bp, k, last_flushed, commit);
-		b = ret ? ERR_PTR(ret) : NULL;
-		goto err;
+		return ret ? ERR_PTR(ret) : NULL;
 	}
 
 	BUG_ON(b->c.level != bp.v->level - 1);
@@ -295,22 +293,19 @@ static struct btree *__bch2_backpointer_get_node(struct btree_trans *trans,
 		return b;
 
 	if (btree_node_will_make_reachable(b)) {
-		b = ERR_PTR(bch_err_throw(c, backpointer_to_overwritten_btree_node));
+		return ERR_PTR(bch_err_throw(c, backpointer_to_overwritten_btree_node));
 	} else {
 		int ret = backpointer_target_not_found(trans, bp, bkey_i_to_s_c(&b->key),
 						       last_flushed, commit);
-		b = ret ? ERR_PTR(ret) : NULL;
+		return ret ? ERR_PTR(ret) : NULL;
 	}
-err:
-	bch2_trans_iter_exit(iter);
-	return b;
 }
 
 static struct bkey_s_c __bch2_backpointer_get_key(struct btree_trans *trans,
 						  struct bkey_s_c_backpointer bp,
 						  struct btree_iter *iter,
 						  unsigned iter_flags,
-						  struct bkey_buf *last_flushed,
+						  struct wb_maybe_flush *last_flushed,
 						  bool commit)
 {
 	struct bch_fs *c = trans->c;
@@ -325,10 +320,8 @@ static struct bkey_s_c __bch2_backpointer_get_key(struct btree_trans *trans,
 				  bp.v->level,
 				  iter_flags);
 	struct bkey_s_c k = bch2_btree_iter_peek_slot(iter);
-	if (bkey_err(k)) {
-		bch2_trans_iter_exit(iter);
+	if (bkey_err(k))
 		return k;
-	}
 
 	/*
 	 * peek_slot() doesn't normally return NULL - except when we ask for a
@@ -346,8 +339,6 @@ static struct bkey_s_c __bch2_backpointer_get_key(struct btree_trans *trans,
 	    extent_matches_bp(c, bp.v->btree_id, bp.v->level, k, bp))
 		return k;
 
-	bch2_trans_iter_exit(iter);
-
 	if (!bp.v->level) {
 		int ret = backpointer_target_not_found(trans, bp, k, last_flushed, commit);
 		return ret ? bkey_s_c_err(ret) : bkey_s_c_null;
@@ -365,7 +356,7 @@ static struct bkey_s_c __bch2_backpointer_get_key(struct btree_trans *trans,
 struct btree *bch2_backpointer_get_node(struct btree_trans *trans,
 					struct bkey_s_c_backpointer bp,
 					struct btree_iter *iter,
-					struct bkey_buf *last_flushed)
+					struct wb_maybe_flush *last_flushed)
 {
 	return __bch2_backpointer_get_node(trans, bp, iter, last_flushed, true);
 }
@@ -374,13 +365,13 @@ struct bkey_s_c bch2_backpointer_get_key(struct btree_trans *trans,
 					 struct bkey_s_c_backpointer bp,
 					 struct btree_iter *iter,
 					 unsigned iter_flags,
-					 struct bkey_buf *last_flushed)
+					 struct wb_maybe_flush *last_flushed)
 {
 	return __bch2_backpointer_get_key(trans, bp, iter, iter_flags, last_flushed, true);
 }
 
 static int bch2_check_backpointer_has_valid_bucket(struct btree_trans *trans, struct bkey_s_c k,
-						   struct bkey_buf *last_flushed)
+						   struct wb_maybe_flush *last_flushed)
 {
 	if (k.k->type != KEY_TYPE_backpointer)
 		return 0;
@@ -422,25 +413,24 @@ fsck_err:
 int bch2_check_btree_backpointers(struct bch_fs *c)
 {
 	struct progress_indicator_state progress;
-
 	bch2_progress_init(&progress, c, BIT_ULL(BTREE_ID_backpointers));
 
-	struct bkey_buf last_flushed __cleanup(bch2_bkey_buf_exit);
-	bch2_bkey_buf_init(&last_flushed);
+	struct wb_maybe_flush last_flushed __cleanup(wb_maybe_flush_exit);
+	wb_maybe_flush_init(&last_flushed);
 
 	CLASS(btree_trans, trans)(c);
 	return for_each_btree_key_commit(trans, iter,
 			BTREE_ID_backpointers, POS_MIN, 0, k,
 			NULL, NULL, BCH_TRANS_COMMIT_no_enospc, ({
-		progress_update_iter(trans, &progress, &iter);
+		progress_update_iter(trans, &progress, &iter) ?:
 		bch2_check_backpointer_has_valid_bucket(trans, k, &last_flushed);
 	}));
 }
 
 struct extents_to_bp_state {
-	struct bpos	bp_start;
-	struct bpos	bp_end;
-	struct bkey_buf last_flushed;
+	struct bpos		bp_start;
+	struct bpos		bp_end;
+	struct wb_maybe_flush	last_flushed;
 };
 
 static int drop_dev_and_update(struct btree_trans *trans, enum btree_id btree,
@@ -522,127 +512,130 @@ err:
 	return ret;
 }
 
-static int check_bp_exists(struct btree_trans *trans,
-			   struct extents_to_bp_state *s,
-			   struct bkey_i_backpointer *bp,
-			   struct bkey_s_c orig_k)
+static int bp_missing(struct btree_trans *trans,
+		      struct bkey_s_c extent,
+		      struct bkey_i_backpointer *bp,
+		      struct bkey_s_c bp_found)
 {
 	struct bch_fs *c = trans->c;
-	struct btree_iter other_extent_iter = {};
-	CLASS(printbuf, buf)();
 	int ret = 0;
 
-	CLASS(btree_iter, bp_iter)(trans, BTREE_ID_backpointers, bp->k.p, 0);
-	struct bkey_s_c bp_k = bkey_try(bch2_btree_iter_peek_slot(&bp_iter));
+	CLASS(printbuf, buf)();
+	prt_str(&buf, "missing backpointer\nfor:  ");
+	bch2_bkey_val_to_text(&buf, c, extent);
+	prt_printf(&buf, "\nwant: ");
+	bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&bp->k_i));
 
-	if (bp_k.k->type != KEY_TYPE_backpointer ||
-	    memcmp(bkey_s_c_to_backpointer(bp_k).v, &bp->v, sizeof(bp->v))) {
-		try(bch2_btree_write_buffer_maybe_flush(trans, orig_k, &s->last_flushed));
-		goto check_existing_bp;
+	if (!bkey_deleted(bp_found.k)) {
+		prt_printf(&buf, "\ngot:  ");
+		bch2_bkey_val_to_text(&buf, c, bp_found);
 	}
-out:
-err:
+
+	if (fsck_err(trans, ptr_to_missing_backpointer, "%s", buf.buf))
+		try(bch2_bucket_backpointer_mod(trans, extent, bp, true));
 fsck_err:
-	bch2_trans_iter_exit(&other_extent_iter);
 	return ret;
-check_existing_bp:
-	/* Do we have a backpointer for a different extent? */
-	if (bp_k.k->type != KEY_TYPE_backpointer)
-		goto missing;
+}
 
-	struct bkey_s_c_backpointer other_bp = bkey_s_c_to_backpointer(bp_k);
+static bool bkey_dev_ptr_stale(struct bch_fs *c, struct bkey_s_c k, unsigned dev)
+{
+	guard(rcu)();
+	struct bch_dev *ca = bch2_dev_rcu_noerror(c, dev);
+	if (!ca)
+		return false;
 
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	bkey_for_each_ptr(ptrs, ptr)
+		if (ptr->dev == dev &&
+		    dev_ptr_stale_rcu(ca, ptr))
+			return true;
+	return false;
+}
+
+static int check_bp_dup(struct btree_trans *trans,
+			struct extents_to_bp_state *s,
+			struct bkey_s_c extent,
+			struct bkey_i_backpointer *bp,
+			struct bkey_s_c_backpointer other_bp)
+{
+	struct bch_fs *c = trans->c;
+
+	CLASS(btree_iter_uninit, other_extent_iter)(trans);
 	struct bkey_s_c other_extent =
 		__bch2_backpointer_get_key(trans, other_bp, &other_extent_iter, 0, NULL, false);
-	ret = bkey_err(other_extent);
+	int ret = bkey_err(other_extent);
+
 	if (ret == -BCH_ERR_backpointer_to_overwritten_btree_node)
-		ret = 0;
+		return bp_missing(trans, extent, bp, other_bp.s_c);
 	if (ret)
-		goto err;
-
+		return ret;
 	if (!other_extent.k)
-		goto missing;
+		return bp_missing(trans, extent, bp, other_bp.s_c);
 
-	rcu_read_lock();
-	struct bch_dev *ca = bch2_dev_rcu_noerror(c, bp->k.p.inode);
-	if (ca) {
-		struct bkey_ptrs_c other_extent_ptrs = bch2_bkey_ptrs_c(other_extent);
-		bkey_for_each_ptr(other_extent_ptrs, ptr)
-			if (ptr->dev == bp->k.p.inode &&
-			    dev_ptr_stale_rcu(ca, ptr)) {
-				rcu_read_unlock();
-				ret = drop_dev_and_update(trans, other_bp.v->btree_id,
-							  other_extent, bp->k.p.inode);
-				if (ret)
-					goto err;
-				goto out;
-			}
+	if (bkey_dev_ptr_stale(c, other_extent, bp->k.p.inode)) {
+		try(drop_dev_and_update(trans, other_bp.v->btree_id, other_extent, bp->k.p.inode));
+		return 0;
 	}
-	rcu_read_unlock();
 
-	if (bch2_extents_match(orig_k, other_extent)) {
-		printbuf_reset(&buf);
+	if (bch2_extents_match(extent, other_extent)) {
+		CLASS(printbuf, buf)();
 		prt_printf(&buf, "duplicate versions of same extent, deleting smaller\n");
-		bch2_bkey_val_to_text(&buf, c, orig_k);
+		bch2_bkey_val_to_text(&buf, c, extent);
 		prt_newline(&buf);
 		bch2_bkey_val_to_text(&buf, c, other_extent);
 		bch_err(c, "%s", buf.buf);
 
-		if (other_extent.k->size <= orig_k.k->size) {
-			ret = drop_dev_and_update(trans, other_bp.v->btree_id,
-						  other_extent, bp->k.p.inode);
-			if (ret)
-				goto err;
-			goto out;
+		if (other_extent.k->size <= extent.k->size) {
+			try(drop_dev_and_update(trans, other_bp.v->btree_id, other_extent, bp->k.p.inode));
+			return 0;
 		} else {
-			ret = drop_dev_and_update(trans, bp->v.btree_id, orig_k, bp->k.p.inode);
-			if (ret)
-				goto err;
-			goto missing;
+			try(drop_dev_and_update(trans, bp->v.btree_id, extent, bp->k.p.inode));
+			return bp_missing(trans, extent, bp, other_bp.s_c);
 		}
+	} else {
+		ret = check_extent_checksum(trans,
+					    other_bp.v->btree_id, other_extent,
+					    bp->v.btree_id, extent,
+					    bp->k.p.inode);
+		if (ret < 0)
+			return ret;
+		if (ret)
+			return bp_missing(trans, extent, bp, other_bp.s_c);
+
+		ret = check_extent_checksum(trans, bp->v.btree_id, extent,
+					    other_bp.v->btree_id, other_extent, bp->k.p.inode);
+		if (ret < 0)
+			return ret;
+		if (ret)
+			return 0;
+
+		CLASS(printbuf, buf)();
+		prt_printf(&buf, "duplicate extents pointing to same space on dev %llu\n", bp->k.p.inode);
+		bch2_bkey_val_to_text(&buf, c, extent);
+		prt_newline(&buf);
+		bch2_bkey_val_to_text(&buf, c, other_extent);
+		bch_err(c, "%s", buf.buf);
+		return bch_err_throw(c, fsck_repair_unimplemented);
+	}
+}
+
+static int check_bp_exists(struct btree_trans *trans,
+			   struct extents_to_bp_state *s,
+			   struct bkey_s_c extent,
+			   struct bkey_i_backpointer *bp)
+{
+	CLASS(btree_iter, bp_iter)(trans, BTREE_ID_backpointers, bp->k.p, 0);
+	struct bkey_s_c bp_found = bkey_try(bch2_btree_iter_peek_slot(&bp_iter));
+
+	if (bp_found.k->type != KEY_TYPE_backpointer) {
+		try(bch2_btree_write_buffer_maybe_flush(trans, extent, &s->last_flushed));
+		try(bp_missing(trans, extent, bp, bp_found));
+	} else if (memcmp(bkey_s_c_to_backpointer(bp_found).v, &bp->v, sizeof(bp->v))) {
+		try(bch2_btree_write_buffer_maybe_flush(trans, extent, &s->last_flushed));
+		try(check_bp_dup(trans, s, extent, bp, bkey_s_c_to_backpointer(bp_found)));
 	}
 
-	ret = check_extent_checksum(trans,
-				    other_bp.v->btree_id, other_extent,
-				    bp->v.btree_id, orig_k,
-				    bp->k.p.inode);
-	if (ret < 0)
-		goto err;
-	if (ret) {
-		ret = 0;
-		goto missing;
-	}
-
-	ret = check_extent_checksum(trans, bp->v.btree_id, orig_k,
-				    other_bp.v->btree_id, other_extent, bp->k.p.inode);
-	if (ret < 0)
-		goto err;
-	if (ret) {
-		ret = 0;
-		goto out;
-	}
-
-	printbuf_reset(&buf);
-	prt_printf(&buf, "duplicate extents pointing to same space on dev %llu\n", bp->k.p.inode);
-	bch2_bkey_val_to_text(&buf, c, orig_k);
-	prt_newline(&buf);
-	bch2_bkey_val_to_text(&buf, c, other_extent);
-	bch_err(c, "%s", buf.buf);
-	ret = bch_err_throw(c, fsck_repair_unimplemented);
-	goto err;
-missing:
-	printbuf_reset(&buf);
-	prt_str(&buf, "missing backpointer\nfor:  ");
-	bch2_bkey_val_to_text(&buf, c, orig_k);
-	prt_printf(&buf, "\nwant: ");
-	bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&bp->k_i));
-	prt_printf(&buf, "\ngot:  ");
-	bch2_bkey_val_to_text(&buf, c, bp_k);
-
-	if (fsck_err(trans, ptr_to_missing_backpointer, "%s", buf.buf))
-		ret = bch2_bucket_backpointer_mod(trans, orig_k, bp, true);
-
-	goto out;
+	return 0;
 }
 
 static int check_extent_to_backpointers(struct btree_trans *trans,
@@ -685,7 +678,7 @@ static int check_extent_to_backpointers(struct btree_trans *trans,
 			continue;
 
 		try(!empty
-		    ? check_bp_exists(trans, s, &bp, k)
+		    ? check_bp_exists(trans, s, k, &bp)
 		    : bch2_bucket_backpointer_mod(trans, k, &bp, true));
 	}
 
@@ -770,13 +763,6 @@ static int bch2_get_btree_in_memory_pos(struct btree_trans *trans,
 	return ret;
 }
 
-static inline int bch2_fs_going_ro(struct bch_fs *c)
-{
-	return test_bit(BCH_FS_going_ro, &c->flags)
-		? -EROFS
-		: 0;
-}
-
 static int bch2_check_extents_to_backpointers_pass(struct btree_trans *trans,
 						   struct extents_to_bp_state *s)
 {
@@ -800,8 +786,8 @@ static int bch2_check_extents_to_backpointers_pass(struct btree_trans *trans,
 			CLASS(btree_node_iter, iter)(trans, btree_id, POS_MIN, 0, level, BTREE_ITER_prefetch);
 
 			try(for_each_btree_key_continue(trans, iter, 0, k, ({
-				bch2_progress_update_iter(trans, &progress, &iter, "extents_to_backpointers");
-				bch2_fs_going_ro(c) ?:
+				bch2_progress_update_iter(trans, &progress, &iter, "extents_to_backpointers") ?:
+				wb_maybe_flush_inc(&s->last_flushed) ?:
 				check_extent_to_backpointers(trans, s, btree_id, level, k) ?:
 				bch2_trans_commit(trans, NULL, NULL, BCH_TRANS_COMMIT_no_enospc);
 			})));
@@ -837,11 +823,11 @@ static int data_type_to_alloc_counter(enum bch_data_type t)
 }
 
 static int check_bucket_backpointers_to_extents(struct btree_trans *, struct bch_dev *, struct bpos,
-						struct bkey_buf *last_flushed);
+						struct wb_maybe_flush *last_flushed);
 
 static int check_bucket_backpointer_mismatch(struct btree_trans *trans, struct bkey_s_c alloc_k,
 					     bool *had_mismatch,
-					     struct bkey_buf *last_flushed,
+					     struct wb_maybe_flush *last_flushed,
 					     struct bpos *last_pos,
 					     unsigned *nr_iters)
 {
@@ -1112,12 +1098,12 @@ int bch2_check_extents_to_backpointers(struct bch_fs *c)
 	struct bpos last_pos = POS_MIN;
 	unsigned nr_iters = 0;
 
-	bch2_bkey_buf_init(&s.last_flushed);
+	wb_maybe_flush_init(&s.last_flushed);
 
 	ret = for_each_btree_key(trans, iter, BTREE_ID_alloc,
 				 POS_MIN, BTREE_ITER_prefetch, k, ({
 		bool had_mismatch;
-		bch2_fs_going_ro(c) ?:
+		bch2_recovery_cancelled(c) ?:
 		check_bucket_backpointer_mismatch(trans, k, &had_mismatch, &s.last_flushed,
 						  &last_pos, &nr_iters);
 	}));
@@ -1144,8 +1130,8 @@ int bch2_check_extents_to_backpointers(struct bch_fs *c)
 
 		if ( bpos_eq(s.bp_start, POS_MIN) &&
 		    !bpos_eq(s.bp_end, SPOS_MAX))
-			bch_verbose(c, "%s(): alloc info does not fit in ram, running in multiple passes with %zu nodes per pass",
-				    __func__, btree_nodes_fit_in_ram(c));
+			bch_info(c, "%s(): alloc info does not fit in ram, running in multiple passes with %zu nodes per pass",
+				 __func__, btree_nodes_fit_in_ram(c));
 
 		if (!bpos_eq(s.bp_start, POS_MIN) ||
 		    !bpos_eq(s.bp_end, SPOS_MAX)) {
@@ -1171,7 +1157,7 @@ int bch2_check_extents_to_backpointers(struct bch_fs *c)
 		bch2_bucket_bitmap_free(&ca->bucket_backpointer_empty);
 	}
 err:
-	bch2_bkey_buf_exit(&s.last_flushed);
+	wb_maybe_flush_exit(&s.last_flushed);
 	bch2_btree_cache_unpin(c);
 	return ret;
 }
@@ -1179,7 +1165,7 @@ err:
 static int check_bucket_backpointer_pos_mismatch(struct btree_trans *trans,
 						 struct bpos bucket,
 						 bool *had_mismatch,
-						 struct bkey_buf *last_flushed)
+						 struct wb_maybe_flush *last_flushed)
 {
 	CLASS(btree_iter, alloc_iter)(trans, BTREE_ID_alloc, bucket, BTREE_ITER_cached);
 	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_slot(&alloc_iter));
@@ -1194,7 +1180,7 @@ static int check_bucket_backpointer_pos_mismatch(struct btree_trans *trans,
 int bch2_check_bucket_backpointer_mismatch(struct btree_trans *trans,
 					   struct bch_dev *ca, u64 bucket,
 					   bool copygc,
-					   struct bkey_buf *last_flushed)
+					   struct wb_maybe_flush *last_flushed)
 {
 	struct bch_fs *c = trans->c;
 	bool had_mismatch;
@@ -1227,7 +1213,7 @@ static int check_one_backpointer(struct btree_trans *trans,
 				 struct bbpos start,
 				 struct bbpos end,
 				 struct bkey_s_c bp_k,
-				 struct bkey_buf *last_flushed)
+				 struct wb_maybe_flush *last_flushed)
 {
 	if (bp_k.k->type != KEY_TYPE_backpointer)
 		return 0;
@@ -1249,7 +1235,7 @@ static int check_one_backpointer(struct btree_trans *trans,
 
 static int check_bucket_backpointers_to_extents(struct btree_trans *trans,
 						struct bch_dev *ca, struct bpos bucket,
-						struct bkey_buf *last_flushed)
+						struct wb_maybe_flush *last_flushed)
 {
 	u32 restart_count = trans->restart_count;
 
@@ -1269,15 +1255,15 @@ static int bch2_check_backpointers_to_extents_pass(struct btree_trans *trans,
 						   struct bbpos start,
 						   struct bbpos end)
 {
-	struct bkey_buf last_flushed __cleanup(bch2_bkey_buf_exit);
-	bch2_bkey_buf_init(&last_flushed);
+	struct wb_maybe_flush last_flushed __cleanup(wb_maybe_flush_exit);
+	wb_maybe_flush_init(&last_flushed);
 
 	struct progress_indicator_state progress;
 	bch2_progress_init(&progress, trans->c, BIT_ULL(BTREE_ID_backpointers));
 
 	return for_each_btree_key(trans, iter, BTREE_ID_backpointers,
 				     POS_MIN, BTREE_ITER_prefetch, k, ({
-			bch2_progress_update_iter(trans, &progress, &iter, "backpointers_to_extents");
+			bch2_progress_update_iter(trans, &progress, &iter, "backpointers_to_extents") ?:
 			check_one_backpointer(trans, start, end, k, &last_flushed);
 	}));
 }
